from CPACqc.services.logging_service import LoggerService
import logging

class ConsoleLogger(LoggerService):
    """
    A simple console logger that implements the LoggerAdapter interface.
    """
    def __init__(self, name: str):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

    def info(self, message: str):
        self.logger.info(message)

    def error(self, message: str):
        self.logger.error(message)

    def debug(self, message: str):
        self.logger.debug(message)

    def warning(self, message: str):
        self.logger.warning(message)

class FileLogger(LoggerService):
    """
    A file logger that implements the LoggerAdapter interface.
    """
    def __init__(self, name: str, filename: str):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        fh = logging.FileHandler(filename)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

    def info(self, message: str):
        self.logger.info(message)

    def error(self, message: str):
        self.logger.error(message)

    def debug(self, message: str):
        self.logger.debug(message)

    def warning(self, message: str):
        self.logger.warning(message)




import pandas as pd
from fnmatch import fnmatch
from CPACqc.core.logger import logger
from reportlab.lib.pagesizes import letter
from reportlab.pdfgen import canvas
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Paragraph, Table, TableStyle
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib import colors
from colorama import Fore, Style
import os
import json
from dataclasses import dataclass, field
from CPACqc.services.report_generator_service import ReportGeneratorService

@dataclass
class Report(ReportGeneratorService):
    # Singleton instance
    _instance: "Report" = field(default=None, init=False, repr=False)

    # Attributes for the Report class
    df: pd.DataFrame = None
    qc_dir: str = None
    sub_ses: str = None
    overlay_df: pd.DataFrame = None
    pdf_path: str = field(init=False)
    pdf_canvas:  canvas.Canvas = field(init=False)  # Renamed to avoid conflict
    width: int = field(init=False, default=letter[0])
    height: int = field(init=False, default=letter[1])
    styles: object = field(init=False, default_factory=getSampleStyleSheet)
    missing_files: list = field(default_factory=list)
    
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Report, cls).__new__(cls)
        return cls._instance

    def __post_init__(self):
        if self.qc_dir and self.sub_ses:
            self.pdf_path = self.get_pdf_path()
            self.pdf_canvas = canvas.Canvas(self.pdf_path, pagesize=letter)
        else:
            self.pdf_path = None
            self.pdf_canvas = None


    @classmethod
    def destroy_instance(cls):
        """Destroy the singleton instance."""
        if cls._instance:
            cls._instance = None
            
    def add_missing_files(self, missing_files):
        self.missing_files.append(missing_files) 
        return

    def get_pdf_path(self):
        pdf = f"{self.sub_ses}_qc_report.pdf"
        base_dir = self.qc_dir if self.qc_dir and ".temp" not in self.qc_dir else os.getcwd()
        return os.path.join(base_dir, pdf)

    def add_front_page(self):
        logo_path = 'https://avatars.githubusercontent.com/u/2230402?s=200&v=4'
        logo_img = ImageReader(logo_path)
        logo_width = 150
        logo_height = 150

        self.pdf_canvas.setFont("Helvetica", 25)
        title = Paragraph(f"{self.sub_ses.replace('_', ' ')}", self.styles['Title'])
        title.wrapOn(self.pdf_canvas, self.width - 40, self.height)
        title.drawOn(self.pdf_canvas, 20, self.height - 100)

        self.pdf_canvas.drawImage(logo_img, (self.width - logo_width) / 2, (self.height - logo_height) / 2, width=logo_width, height=logo_height)
        self.pdf_canvas.setFont("Helvetica", 15)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height - 120, "Quality Control Report")

        self.pdf_canvas.setFont("Helvetica", 12)
        self.pdf_canvas.drawCentredString(self.width / 2, 100, f"Created on: {pd.Timestamp.now().strftime('%Y-%m-%d %H:%M:%S')}")
        self.pdf_canvas.drawCentredString(self.width / 2, 80, "CPAC Team")

        self.pdf_canvas.showPage()

    def add_contents_page(self):
        self.pdf_canvas.setFont("Helvetica", 20)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height - 100, "Contents")
        self.pdf_canvas.setFont("Helvetica", 12)
        y_position = self.height - 150
        chapters = sorted(set(self.df['datatype'].dropna()))
    
        for chapter in chapters:
            if y_position < 50:
                self.pdf_canvas.showPage()
                y_position = self.height - 50
            y_position -= 20
            self.pdf_canvas.setFillColor(colors.blue)
            self.pdf_canvas.drawString(80, y_position, f"{chapter}")
            text_width = self.pdf_canvas.stringWidth(chapter, "Helvetica", 12)
            self.pdf_canvas.linkRect("", f"chapter_{chapter}", (80, y_position - 2, 80 + text_width, y_position + 10), color=colors.blue)
            self.pdf_canvas.bookmarkPage(f"chapter_{chapter}")
            self.pdf_canvas.addOutlineEntry(f"{chapter}", f"chapter_{chapter}", level=0)
            y_position -= 10
    
            chapter_data = self.df[self.df['datatype'] == chapter]
            scans = sorted(set(chapter_data['scan'].dropna()))
            for scan in scans:
                if y_position < 50:
                    self.pdf_canvas.showPage()
                    y_position = self.height - 50
                self.pdf_canvas.setFillColor(colors.green)
                self.pdf_canvas.drawString(100, y_position, f"{scan}")
                text_width = self.pdf_canvas.stringWidth(scan, "Helvetica", 12)
                if scan.strip() == '':
                    self.pdf_canvas.linkRect("", f"subsection_{chapter}", (100, y_position - 2, 100 + text_width, y_position + 10), color=colors.green)
                    self.pdf_canvas.bookmarkPage(f"subsection_{chapter}")
                    self.pdf_canvas.addOutlineEntry(f"{chapter}", f"subsection_{chapter}", level=1)
                else:
                    self.pdf_canvas.linkRect("", f"subsection_{chapter}_{scan}", (100, y_position - 2, 100 + text_width, y_position + 10), color=colors.green)
                    self.pdf_canvas.bookmarkPage(f"subsection_{chapter}_{scan}")
                    self.pdf_canvas.addOutlineEntry(f"{chapter} - {scan}", f"subsection_{chapter}_{scan}", level=1)
                y_position -= 10
    
                scan_data = chapter_data[chapter_data['scan'] == scan]
    
                if not scan_data.empty:
                    ordered_images = []
                    extra_images = []
    
                    if self.overlay_df is not None:
                        ordered_images = scan_data[scan_data['resource_name'].isin(self.overlay_df['output'])]
                        extra_images = scan_data[~scan_data['resource_name'].isin(self.overlay_df['output'])]
                    else:
                        ordered_images = scan_data
    
                    # Remove duplicates, keeping only the last occurrence
                    ordered_images = ordered_images.drop_duplicates(subset='file_name', keep='last')
                    extra_images = extra_images.drop_duplicates(subset='file_name', keep='last')
    
                    for _, image_data in ordered_images.iterrows():
                        if y_position < 50:
                            self.pdf_canvas.showPage()
                            y_position = self.height - 50
                        self.pdf_canvas.setFillColor(colors.black)
                        self.pdf_canvas.drawString(120, y_position, f"{image_data['resource_name']}")
                        text_width = self.pdf_canvas.stringWidth(image_data['resource_name'], "Helvetica", 12)
                        self.pdf_canvas.linkRect("", f"image_{chapter}_{scan}_{image_data['resource_name']}", (120, y_position - 2, 120 + text_width, y_position + 10), color=colors.blue)
                        self.pdf_canvas.bookmarkPage(f"image_{chapter}_{scan}_{image_data['resource_name']}")
                        self.pdf_canvas.addOutlineEntry(f"{image_data['resource_name']}", f"image_{chapter}_{scan}_{image_data['resource_name']}", level=2)
                        y_position -= 13
    
                    for _, image_data in extra_images.iterrows():
                        if y_position < 50:
                            self.pdf_canvas.showPage()
                            y_position = self.height - 50
                        self.pdf_canvas.setFillColor(colors.black)
                        self.pdf_canvas.drawString(120, y_position, f"{image_data['resource_name']}")
                        text_width = self.pdf_canvas.stringWidth(image_data['resource_name'], "Helvetica", 12)
                        self.pdf_canvas.linkRect("", f"image_{chapter}_{scan}_{image_data['resource_name']}", (120, y_position - 2, 120 + text_width, y_position + 10), color=colors.blue)
                        self.pdf_canvas.bookmarkPage(f"image_{chapter}_{scan}_{image_data['resource_name']}")
                        self.pdf_canvas.addOutlineEntry(f"{image_data['resource_name']}", f"image_{chapter}_{scan}_{image_data['resource_name']}", level=2)
                        y_position -= 13
    
                y_position -= 10
        self.pdf_canvas.showPage()

    def add_images(self):
        chapters = sorted(set(self.df['datatype'].dropna()))
        page_number = 1  

        for chapter in chapters:
            self.add_chapter_title_page(chapter)
            page_number += 1

            chapter_data = self.df[self.df['datatype'] == chapter]
            scans = sorted(set(chapter_data['scan'].dropna()))

            for scan in scans:

                self.add_scan_title_page(chapter, scan)
                page_number += 1

                scan_data = chapter_data[chapter_data['scan'] == scan]

                if not scan_data.empty:
                    ordered_images = []
                    extra_images = []

                    if self.overlay_df is not None:
                        ordered_images = scan_data[scan_data['resource_name'].isin(self.overlay_df['output'])]
                        extra_images = scan_data[~scan_data['resource_name'].isin(self.overlay_df['output'])]
                    else:
                        ordered_images = scan_data

                    # Remove duplicates, keeping only the last occurrence
                    ordered_images = ordered_images.drop_duplicates(subset='file_name', keep='last')
                    extra_images = extra_images.drop_duplicates(subset='file_name', keep='last')

                    y_position = self.height - 30

                    for _, image_data in ordered_images.iterrows():
                        y_position, page_number = self.add_image(image_data, chapter, scan, y_position, page_number)

                    for _, image_data in extra_images.iterrows():
                        y_position, page_number = self.add_image(image_data, chapter, scan, y_position, page_number)

    def add_chapter_title_page(self, chapter):
        self.pdf_canvas.setFont("Helvetica", 30)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height / 2, chapter)
        self.pdf_canvas.bookmarkPage(f"chapter_{chapter}")
        self.pdf_canvas.showPage()

    def add_scan_title_page(self, chapter, scan):
        if scan.strip() == '':
            print(Fore.YELLOW + f"Scan name is empty for chapter {chapter}. Skipping..." + Style.RESET_ALL)
            self.pdf_canvas.bookmarkPage(f"subsection_{chapter}")
            return
        self.pdf_canvas.setFont("Helvetica", 25)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height / 2, f"{chapter} - {scan}")
        self.pdf_canvas.bookmarkPage(f"subsection_{chapter}_{scan}")
        self.pdf_canvas.showPage()

    def add_image(self, image_data, chapter, scan, y_position, page_number):
        image_path = os.path.join(self.qc_dir, image_data['relative_path'])
        if os.path.exists(image_path):
            img = ImageReader(image_path)
            max_img_width = self.width - 20
            max_img_height = self.height - 100
    
            img_width, img_height = img.getSize()
            aspect_ratio = img_width / img_height
            if aspect_ratio > 1:
                img_width = max_img_width
                img_height = img_width / aspect_ratio
            else:
                img_height = max_img_height
                img_width = img_height * aspect_ratio
    
            if y_position - img_height - 140 < 0:
                self.add_footer(page_number)
                page_number += 1
                y_position = self.height - 30
                self.add_header(chapter, scan)
    
            self.pdf_canvas.setFont("Helvetica", 15)
            self.pdf_canvas.drawString(10, y_position - 20, f"{image_data['resource_name']}")
            self.pdf_canvas.bookmarkPage(f"image_{chapter}_{scan}_{image_data['resource_name']}")
    
            self.pdf_canvas.drawImage(img, (self.width - img_width) / 2, y_position - img_height - 40, width=img_width, height=img_height)
    
            # Draw image details table
            table_height = self.add_image_details(image_data, y_position, img_height)
    
            # Draw JSON details table below image details
            json_table_height = self.add_json_details(image_data, y_position - img_height - table_height - 80)
    
            y_position -= img_height + table_height + json_table_height + 90
    
            self.pdf_canvas.showPage()
        return y_position, page_number
    
    def add_image_details(self, image_data, y_position, img_height):

        label = f"{image_data['file_name']}"
        self.styles['Normal'].textColor = colors.whitesmoke
        wrapped_label = Paragraph(label, self.styles['Normal'])
        wrapped_label.wrapOn(self.pdf_canvas, self.width - 20, self.height)
    
        file_info = json.loads(image_data['file_info'])
        file_info_text = [
            ["Image:", wrapped_label],
            ["Orientation:", file_info['orientation']],
            ["Dimensions:", " x ".join(map(str, file_info['dimension']))],
            ["Resolution:", " x ".join(map(lambda x: str(round(x, 2)), file_info['resolution']))]
        ]
    
        if file_info['tr'] is not None:
            file_info_text.append(["TR:", str(round(file_info['tr'], 2))])
    
        if file_info['nos_tr'] is not None:
            file_info_text.append(["No of TRs:", str(file_info['nos_tr'])])
    
        table = Table(file_info_text, colWidths=[80, self.width - 100])
        table.setStyle(TableStyle([
            ('BACKGROUND', (0, 0), (-1, 0), colors.darkgreen),
            ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
            ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
            ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
            ('FONTSIZE', (0, 0), (-1, -1), 10),
            ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
            ('BACKGROUND', (0, 1), (-1, -1), colors.lightgrey),
            ('TEXTCOLOR', (0, 1), (-1, -1), colors.black),
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
        ]))
        table.wrapOn(self.pdf_canvas, self.width - 20, self.height)
        table_height = table.wrap(self.width - 20, self.height)[1]
        table.drawOn(self.pdf_canvas, 10, y_position - img_height - table_height - 60)
        return table_height
    
    def add_json_details(self, image_data, y_position):
        flat_json = image_data.get('json', None)
        if flat_json is None or flat_json == '' or flat_json == {}:
            print(Fore.YELLOW + f"JSON data is empty for {image_data.get('file_name', '')}. Skipping..." + Style.RESET_ALL)
            return 0
    
        # If already a dict, skip json.loads
        if isinstance(flat_json, str):
            try:
                flat_json = json.loads(flat_json)
            except Exception as e:
                print(Fore.YELLOW + f"Could not parse JSON for {image_data.get('file_name', '')}: {e}" + Style.RESET_ALL)
                return 0
    
        if not flat_json or flat_json == {}:
            print(Fore.YELLOW + f"JSON data is empty for {image_data.get('file_name', '')}. Skipping..." + Style.RESET_ALL)
            return 0
    
        # Add header row
        json_text = [["Json Field", "Value"]]
        for key, value in flat_json.items():
            # Skip empty keys or NaN values
            if not key or (isinstance(value, float) and pd.isna(value)):
                continue
            # Ensure value text is black
            self.styles['Normal'].textColor = colors.black
            para_key = Paragraph(str(key), self.styles['Normal'])
            para_value = Paragraph(str(value), self.styles['Normal'])
            json_text.append([para_key, para_value])
    
        if len(json_text) == 1:
            print(Fore.YELLOW + f"JSON data is empty for {image_data.get('file_name', '')}. Skipping..." + Style.RESET_ALL)
            return 0
    
        table = Table(json_text, colWidths=[80, self.width - 100])
        table.setStyle(TableStyle([
            ('BACKGROUND', (0, 0), (-1, 0), colors.darkblue),
            ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
            ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
            ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
            ('FONTSIZE', (0, 0), (-1, -1), 10),
            ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
            ('BACKGROUND', (0, 1), (-1, -1), colors.lightgrey),
            ('TEXTCOLOR', (0, 1), (-1, -1), colors.black),
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
        ]))
        table.wrapOn(self.pdf_canvas, self.width - 20, self.height)
        table_height = table.wrap(self.width - 20, self.height)[1]
        table.drawOn(self.pdf_canvas, 10, y_position - table_height)
        return table_height + 10  # Add 10 for spacing


    def add_header(self, chapter, scan):
        self.pdf_canvas.setFont("Helvetica", 12)
        self.pdf_canvas.drawString(10, self.height - 30, f"Chapter : {chapter}/{scan}")

    def add_footer(self, page_number):
        self.pdf_canvas.drawRightString(self.width - 30, 20, str(page_number))

    def add_missing_files_page(self):
        if not self.missing_files:
            print(Fore.YELLOW + "No missing files to report." + Style.RESET_ALL)
            return
        self.pdf_canvas.bookmarkPage("Missing Files")
        self.pdf_canvas.setFont("Helvetica", 20)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height - 120, "Missing Files")
        self.pdf_canvas.setFont("Helvetica", 12)
    
        y_position = self.height - 150
        for file in self.missing_files:
            print(Fore.YELLOW + f"Missing file: {file}" + Style.RESET_ALL)
            if y_position < 50:
                self.pdf_canvas.showPage()
                y_position = self.height - 50
    
            self.pdf_canvas.setFont("Helvetica", 12)
            self.pdf_canvas.drawString(70, y_position - 20, file)
            y_position -= 13
    
        self.pdf_canvas.showPage()
        self.pdf_canvas.setFont("Helvetica", 25)
        self.pdf_canvas.drawCentredString(self.width / 2, self.height / 2, "End of Report")
        self.pdf_canvas.showPage()

    def generate_report(self):
        print(Fore.YELLOW + "Generating PDF report..." + Style.RESET_ALL)
        self.add_front_page()
        self.add_contents_page()
        self.add_images()
        self.add_missing_files_page()
        print(Fore.GREEN + "PDF report generated successfully." + Style.RESET_ALL)

    def save_report(self):
        if self.pdf_canvas:
            self.pdf_canvas.save()
            print(Fore.GREEN + f"PDF report saved at {self.pdf_path}" + Style.RESET_ALL)
        else:
            print(Fore.RED + "PDF canvas is not initialized." + Style.RESET_ALL)




import pandas as pd
import os
import argparse
from multiprocessing import Pool
from tqdm import tqdm
from functools import lru_cache
import tempfile
import bids2table as b2t

import nibabel as nib
from nibabel.orientations import io_orientation, ornt2axcodes
from colorama import Fore, Style, init

import json
import re
from fnmatch import fnmatch

import ast
from CPACqc.core.logger import logger


def make_dir(path) -> None: 
    """
    Create a directory if it doesn't exist.
    
    Parameters:
    path (str): The path to the directory to create.

    Returns:
    pathlib.Path: The created or existing directory.
    """
    path = pathlib.Path(path)
    try:
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        message = f"Error creating directory {path}: {e}"
        logger.error(message=message)
        return None


def get_file_info(file_path):
    try:
        img = nib.load(file_path)
        resolution = tuple(float(x) for x in img.header.get_zooms())
        dimension = tuple(int(x) for x in img.shape)
        
        affine = img.affine
        orientation = "".join(ornt2axcodes(io_orientation(affine))) + " @nibabel"
        
        if len(dimension) == 4:
            # get TR info
            tr = float(img.header.get_zooms()[3])
            nos_tr = str(int(img.shape[-1]))
        else:
            tr = None
            nos_tr = None

        return json.dumps({
            "resolution": resolution,
            "dimension": dimension,
            "tr": tr,
            "nos_tr": nos_tr,
            "orientation": orientation
        })
    except Exception as e:
        message = f"Error processing file {file_path}: {e}"
        logger.error(message=message)
        return None

def gen_resource_name(row):
    sub_ses = row["sub_ses"]

    # Handle task, run, acq flexibly
    task = row.get("task", "") or ""
    run = row.get("run", "") or ""
    acq = row.get("acq", "") or ""

    # Build scan pattern based on available fields
    scan_parts = []
    if task:
        scan_parts.append(f"task-{task}")
    if acq:
        scan_parts.append(f"acq-{acq}")
    if run:
        try:
            scan_parts.append(f"run-{str(int(float(run)))}")
        except ValueError:
            scan_parts.append(f"run-{run}")

    scan_pattern = "_".join(scan_parts)
    if scan_pattern:
        scan_pattern += "_"

    # Build regex pattern to remove from file_name
    pattern = f"{sub_ses}_" + scan_pattern
    resource_name = row["file_name"].replace(pattern, "")
    resource_name = resource_name.replace(f"{sub_ses}_", "")
    return resource_name

def get_rows_by_resource_name(resource_name, datatype, nii_gz_files):
    # Ensure nii_gz_files is a DataFrame and access the correct column
    if isinstance(nii_gz_files, pd.DataFrame):
        # Filter rows using the fnmatch pattern and datatype match if datatype is provided
        if datatype:
            rows = nii_gz_files[
                nii_gz_files['resource_name'].apply(lambda x: fnmatch(x, resource_name)) &
                (nii_gz_files['datatype'] == datatype)
            ]
        else:
            rows = nii_gz_files[
                nii_gz_files['resource_name'].apply(lambda x: fnmatch(x, resource_name))
            ]
        
        if len(rows) == 0:
            message = f"NOT FOUND: {resource_name} with datatype: {datatype}"
            logger.error(message=message)
            return None
        return rows
    else:
        message = "nii_gz_files is not a DataFrame"
        logger.error(message=message)
        return None

# check file_path and drop the ones that are higher dimensions for now
def is_3d_or_4d(file_path):
    dim = len(nib.load(file_path).shape)
    if dim > 4:
        file_name = os.path.basename(file_path).split(".")[0]
        message = f"NOT 3D: {file_name} \n its {dim}D"
        logger.error(message=message)
        message = f"Skipping for now ...."
        logger.error(message=message)

        return False
    return True

def get_scan(row):
    task = row.get("task", "")
    acq = row.get("acq", "")
    run = row.get("run", "")

    # Handle missing or NaN values
    task_str = f"task-{task}" if task and str(task).strip() and str(task).lower() != "nan" else ""
    acq_str = f"acq-{acq}" if acq and str(acq).strip() and str(acq).lower() != "nan" else ""

    try:
        run_str = f"run-{str(int(float(run)))}" if run and str(run).strip() and str(run).lower() != "nan" else ""
        
    except ValueError:
        run_str = f"run-{run}" if run and str(run).strip() and str(run).lower() != "nan" else ""

    # Combine non-empty parts
    parts = [p for p in [task_str, acq_str, run_str] if p]
    return "_".join(parts)

def get_sub_ses(row):
    sub = row["sub"]
    ses = row["ses"] if row["ses"] != "" else False
    return f"sub-{sub}_ses-{ses}" if ses else f"sub-{sub}"

def gen_filename(res1_row, res2_row=None):
    scan = res1_row["scan"] + "_"
    if res2_row is not None:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']} overlaid on {res2_row['resource_name']}"
    else:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']}"

def create_directory(sub, ses, base_dir):
    sub_dir = os.path.join(base_dir, sub, ses)
    os.makedirs(sub_dir, exist_ok=True)
    return sub_dir

def generate_plot_path(sub_dir, file_name):
    return os.path.join(sub_dir, f"{file_name}.png")

def process_row(row, nii_gz_files, overlay_dir, plots_dir, report):
    image_1 = row.get("output", False)
    image_2 = row.get("underlay", False)
    datatype = row.get("datatype", False)

    resource_name_1 = get_rows_by_resource_name(image_1, datatype, nii_gz_files) if image_1 else None
    resource_name_2 = get_rows_by_resource_name(image_2, False, nii_gz_files) if image_2 else None

    if resource_name_1 is None:
        message = f"NOT FOUND: {image_1}"
        logger.error(message=message)
        report.add_missing_files(image_1)
        return []

    result_rows = []
    seen = set()  # To track duplicates

    for _, res1_row in resource_name_1.iterrows():
        if resource_name_2 is not None:
            result_rows.extend(process_res2_rows(res1_row, resource_name_2, seen, overlay_dir, plots_dir))
        else:
            result_rows.extend(process_single_row(res1_row, seen, plots_dir))

    return result_rows

def process_res2_rows(res1_row, resource_name_2, seen, overlay_dir, plots_dir):
    result_rows = []
    for _, res2_row in resource_name_2.iterrows():
        file_name = gen_filename(res1_row, res2_row)
        if res1_row['space'] == res2_row['space']:
            if file_name not in seen:
                seen.add(file_name)
                sub_dir = create_directory(res1_row['sub'], res1_row['ses'], overlay_dir)
                plot_path = generate_plot_path(sub_dir, file_name)
                result_rows.append(create_result_row(res1_row, res2_row, file_name, overlay_dir, plot_path))
        elif res1_row['space'] == "" or res2_row['space'] == "":
            pass
        else:
            message = f"SPACE MISMATCH while trying to plot .. \n{file_name.split('overlaid on')[0]}\nover\n{file_name.split(' overlaid on ')[1]}\n{res1_row['space']} != {res2_row['space']}"
            logger.info(message=message)

    if len(result_rows) == 0:
        message = f"MATCHING SPACE NOT-FOUND to plot {res1_row['resource_name']} over {resource_name_2['resource_name'].tolist()}"
        logger.error(message=message)
        file_name = gen_filename(res1_row) + " *SPACE MISMATCH TO OVERLAY!*"
        if file_name not in seen:
            seen.add(file_name)
            sub_dir = create_directory(res1_row['sub'], res1_row['ses'], plots_dir)
            plot_path = generate_plot_path(sub_dir, file_name)
            result_rows.append(create_result_row(res1_row, None, file_name, plots_dir, plot_path))
    return result_rows

def process_single_row(res1_row, seen, plots_dir):
    result_rows = []
    file_name = gen_filename(res1_row)
    if file_name not in seen:
        seen.add(file_name)
        sub_dir = create_directory(res1_row['sub'], res1_row['ses'], plots_dir)
        plot_path = generate_plot_path(sub_dir, file_name)
        result_rows.append(create_result_row(res1_row, None, file_name, plots_dir, plot_path))
    return result_rows

def create_result_row(res1_row, res2_row, file_name, plots_dir, plot_path):
    return {
        "sub": res1_row["sub"],
        "ses": res1_row["ses"],
        "file_path_1": res1_row["file_path"],
        "file_path_2": res2_row["file_path"] if res2_row is not None else None,
        "file_name": file_name,
        "plots_dir": plots_dir,
        "plot_path": plot_path,
        "datatype": res1_row["datatype"],
        "resource_name": res1_row["resource_name"],
        "space": res1_row["space"],
        "scan": res1_row["scan"],
        "json": res1_row["json"]
    }

def fill_space(row):
    if row["space"] == "":
        if row["datatype"] == "anat":
            return "T1w"
        elif row["datatype"] == "func":
            return "bold"
    return row["space"]



KEYS_TO_REMOVE = {"CpacConfig", "CpacProvenance", "CpacConfigHash"}

def remove_keys(d, keys_to_remove):
    if isinstance(d, dict):
        return {k: remove_keys(v, keys_to_remove)
                for k, v in d.items() if k not in keys_to_remove}
    elif isinstance(d, list):
        return [remove_keys(i, keys_to_remove) for i in d]
    return d

def parse_json_column(val):
    """
    Parse a JSON column value into a cleaned Python dict.
    Handles nested dicts/lists and removes unwanted keys.
    """
    try:
        # Try to parse string to dict/list
        if isinstance(val, str):
            parsed = ast.literal_eval(val)
        else:
            parsed = val
        cleaned = remove_keys(parsed, KEYS_TO_REMOVE)
        return cleaned  # returns a dict or list, not a string
    except Exception as e:
        print(f"Error parsing JSON column: {e}")
        return {}


def flatten_json_collapse_lists(d, parent_key='', sep='-'):
    """
    Flatten a nested dict for display, collapsing lists into comma-separated strings.
    """
    items = []
    if isinstance(d, dict):
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(flatten_json_collapse_lists(v, new_key, sep=sep).items())
            elif isinstance(v, list):
                # Collapse list to comma-separated string if not list of dicts
                if all(not isinstance(i, dict) for i in v):
                    items.append((new_key, ", ".join(str(i) for i in v)))
                else:
                    # If list of dicts, flatten each dict
                    for idx, item in enumerate(v):
                        items.extend(flatten_json_collapse_lists(item, f"{new_key}[{idx}]", sep=sep).items())
            else:
                items.append((new_key, v))
    else:
        items.append((parent_key, d))
    return dict(items)

def clean_and_flatten_json_column(val):
    """
    Parse, clean (remove unwanted keys), and flatten the JSON column value into a 1-level dict,
    collapsing lists into comma-separated strings.
    """
    try:
        cleaned = parse_json_column(val)
        flat = flatten_json_collapse_lists(cleaned)
        return flat  # returns a flat dict
    except Exception as e:
        print(f"Error cleaning and flattening JSON column: {e}")
        return {}



import pandas as pd
import os
from colorama import Fore, Style
import json
import shutil

from CPACqc.core.utils import (
    process_row, gen_filename, generate_plot_path, create_directory, get_file_info
)
from CPACqc.utils.multiprocessing.multiprocessing_utils import Multiprocess
from CPACqc.utils.report.pdf import Report
from CPACqc.core.config import Config
from CPACqc.utils.tabler.bids2table import Bids2TableDetails
from CPACqc.core.entity import Table
from CPACqc.utils.plotting.plot import Plot

class QCPipeline:
    def __init__(self, args):
        self.app_config = Config(
            bids_dir=args.bids_dir,
            qc_dir=args.qc_dir,
            subject_list=args.sub,
            overlay_csv=args.config,
            n_procs=args.n_procs
        )
        self.app_config.plots_dir = os.path.join(self.app_config.qc_dir, "plots")
        self.app_config.overlay_dir = os.path.join(self.app_config.qc_dir, "overlays")
        self.app_config.csv_dir = os.path.join(self.app_config.qc_dir, "csv")
        self.app_config.make_dirs()

        self.logger = self.app_config.logger
        self.bids_parser = Bids2TableDetails(
            base_dir=self.app_config.bids_dir,
            subjects=self.app_config.subject_list,
            workers=self.app_config.n_procs
        )
        self.my_table = Table(self.bids_parser.get_dataframe())
        self.nii_gz_files = self.my_table.processed_table
        self.overlay_df = pd.read_csv(self.app_config.overlay_csv).fillna(False)
        self.not_plotted = []

        self.cols_to_keep = ['sub', 'ses', 'file_path_1', 'file_path_2', 'file_name', 'plots_dir', 'plot_path', 'datatype', 'resource_name', 'space', 'scan']

    def run(self):
        sub_ses_list = self.nii_gz_files["sub_ses"].unique()
        for idx, sub_ses in enumerate(sub_ses_list, 1):
            print(Fore.YELLOW + f"Processing subject {idx}/{len(sub_ses_list)}: {sub_ses}" + Style.RESET_ALL)
            sub_df = self.nii_gz_files[self.nii_gz_files["sub_ses"] == sub_ses]
            self.process_subject(sub_ses, sub_df)
        self.cleanup_qc_dir()
        self.logger.info(f"QC pipeline completed for {len(sub_ses_list)} subjects.")

    def process_subject(self, sub_ses, sub_df):
        report = self.init_report(sub_ses)
        result_df = self.generate_results(sub_df, report)
        result_df = self.add_missing_rows(result_df, sub_df)
        result_df = self.add_additional_columns(result_df)
        result_df = self.remove_duplicates(result_df)
        self.save_results(result_df, sub_ses)
        self.run_plotting(result_df)
        self.generate_report(result_df, report)

    def init_report(self, sub_ses):
        return Report(
            qc_dir=self.app_config.qc_dir,
            sub_ses=sub_ses,
            overlay_df=self.overlay_df
        )

    def generate_results(self, sub_df, report):
        results = self.overlay_df.apply(
            lambda row: process_row(row, sub_df, self.app_config.overlay_dir, self.app_config.plots_dir, report),
            axis=1
        ).tolist()
        results = [item for sublist in results for item in sublist]
        return pd.DataFrame(results)

    def add_missing_rows(self, result_df, sub_df):
        if 'file_path_1' not in result_df.columns:
            result_df['file_path_1'] = None
        missing_rows = sub_df.loc[~sub_df['file_path'].isin(result_df['file_path_1'])].copy()
        if not missing_rows.empty:
            missing_rows['file_path_1'] = missing_rows['file_path']
            missing_rows['file_path_2'] = None
            missing_rows['file_name'] = missing_rows.apply(lambda row: gen_filename(res1_row=row), axis=1)
            missing_rows['plots_dir'] = self.app_config.plots_dir
            missing_rows['plot_path'] = missing_rows.apply(
                lambda row: generate_plot_path(create_directory(row['sub'], row['ses'], row['plots_dir']), row['file_name']),
                axis=1
            )
            cols = ['sub', 'ses', 'file_path_1', 'file_path_2', 'file_name', 'plots_dir', 'plot_path', 'datatype', 'resource_name', 'space', 'scan']
            if 'json' in missing_rows.columns:
                cols.append('json')
            missing_rows = missing_rows[cols].copy()
            result_df = pd.concat([result_df, missing_rows], ignore_index=True)
        return result_df

    def add_additional_columns(self, result_df):
        result_df['relative_path'] = result_df.apply(lambda row: os.path.relpath(row['plot_path'], self.app_config.qc_dir), axis=1)
        result_df['file_info'] = result_df.apply(lambda row: get_file_info(row['file_path_1']), axis=1)
        return result_df

    def save_results(self, result_df, sub_ses):
        result_df_csv_path = os.path.join(self.app_config.csv_dir, f"{sub_ses}_results.csv")
        result_df.to_csv(
            result_df_csv_path,
            mode='a' if os.path.exists(result_df_csv_path) else 'w',
            header=not os.path.exists(result_df_csv_path),
            index=False
        )

    def remove_duplicates(self, result_df):
        subset_cols = [
            "file_path_1", "file_path_2", "file_name", "plots_dir", "plot_path",
            "datatype", "resource_name", "space", "scan"
        ]
        if 'json' in result_df.columns:
            result_df['json'] = result_df['json'].apply(
                lambda x: json.dumps(x, sort_keys=True) if isinstance(x, dict) else (x if pd.isna(x) else str(x))
            )
            subset_cols.append("json")
        return result_df.drop_duplicates(subset=subset_cols, keep="first")

    def run_plotting(self, result_df):
        my_plots = [
            Plot(
                sub=row['sub'],
                ses=row['ses'],
                file_path_1=row['file_path_1'],
                file_path_2=row['file_path_2'],
                file_name=row['file_name'],
                plots_dir=row['plots_dir'],
                plot_path=row['plot_path']
            ) for _, row in result_df.iterrows()
        ]
        overlay_multiprocessor = Multiprocess(Plot.plot_overlay, my_plots, self.app_config.n_procs)
        self.not_plotted += overlay_multiprocessor.run()

    def generate_report(self, result_df, report):
        try:
            report.df = result_df
            report.generate_report()
            report.save_report()
            Report.destroy_instance()
        except Exception as e:
            print(Fore.RED + f"Error generating PDF: {e}" + Style.RESET_ALL)

    def cleanup_qc_dir(self):
        """
        Remove temporary QC directory or unnecessary files after processing.
        """
        qc_dir = self.app_config.qc_dir
        if os.path.exists(qc_dir):
            print(Fore.YELLOW + f"Removing the QC output directory: {qc_dir}" + Style.RESET_ALL)
            shutil.rmtree(qc_dir)
        else:
            print(Fore.YELLOW + f"QC output directory does not exist: {qc_dir}" + Style.RESET_ALL)


from abc import ABC, abstractmethod
from typing import Dict, Any, List, Optional

class PlottingService(ABC):

    @abstractmethod
    def plot_overlay(self, sub, ses, file_path_1, file_path_2, file_name, plots_dir, plot_path):
        """
        Abstract method to plot data.
        
        Parameters:
        - sub: str, subject ID
        - ses: str, session ID
        - file_path_1: str, path to the first file
        - file_path_2: Optional[str], path to the second file (if applicable)
        - file_name: str, name of the output plot file
        - plots_dir: str, directory where plots will be saved
        - plot_path: str, full path for the output plot
        
        Returns:
        - None
        """
        pass



import pytest
from CPACqc.ui.cli import get_args

import sys
import argparse
import pytest
from unittest.mock import patch


def test_get_args_required_only():
    test_args = ["prog", "-d", "/path/to/bids"]
    with patch.object(sys, 'argv', test_args):
        args = get_args()
        assert args.bids_dir == "/path/to/bids"
        assert args.qc_dir is None
        assert args.config is None
        assert args.sub is None
        assert args.n_procs == 8

def test_get_args_all_options():
    test_args = [
        "prog", "-d", "/bids", "-o", "/qc", "-c", "config.csv", 
        "-s", "sub-01", "sub-02", "-n", "4"
    ]
    with patch.object(sys, 'argv', test_args):
        args = get_args()
        assert args.bids_dir == "/bids"
        assert args.qc_dir == "/qc"
        assert args.config == "config.csv"
        assert args.sub == ["sub-01", "sub-02"]
        assert args.n_procs == 4




from dataclasses import dataclass, field
from typing import List, Optional
from datetime import datetime
import os
import pandas

from CPACqc.core.logger import logger

@dataclass
class Config:
    bids_dir: str
    qc_dir: Optional[str] = None
    subject_list: Optional[List[str]] = None
    n_procs: int = 8

    logger = logger 
    overlay_csv: Optional[str] = None

    plots_dir: str = field(init=False)
    overlay_dir: str = field(init=False)
    csv_dir: str = field(init=False)

    def __post_init__(self):
        if not self.overlay_csv:
            self.overlay_csv = os.path.join(os.path.dirname(__file__), "..", "utils", "overlay", "overlay.csv")

        if not self.qc_dir:
            self.qc_dir = os.path.join(os.getcwd(), '.temp_qc')


    def make_dirs(self):
        """
        Create necessary directories for QC output.
        """
        os.makedirs(self.qc_dir, exist_ok=True)
        os.makedirs(self.plots_dir, exist_ok=True)
        os.makedirs(self.overlay_dir, exist_ok=True)
        os.makedirs(self.csv_dir, exist_ok=True)



# CPACqc/cli.py

import argparse
from colorama import Fore, Style
from CPACqc.core.QCPipeline import QCPipeline
from CPACqc.core.validation import validate_args
from CPACqc import __version__, __author__, __email__, __description__

def get_args() -> argparse.Namespace:
    """
    Parse command-line arguments.
    """
    version_str = (
        f"{Fore.GREEN}CPACqc version: {__version__}{Style.RESET_ALL}"
    )
    parser = argparse.ArgumentParser(description="Process BIDS directory and generate QC plots.")
    parser.add_argument("-d", "--bids_dir", required=True, help="Path to the BIDS directory")
    parser.add_argument("-o", "--qc_dir", required=False, help="Path to the QC output directory")
    parser.add_argument("-c", "--config", required=False, help="Config file")
    parser.add_argument("-s", "--sub", nargs='+', required=False, help="Specify subject/participant label(s) to process")
    parser.add_argument("-n", "--n_procs", type=int, default=8, help="Number of processes to use for multiprocessing")
    parser.add_argument("-v", "--version", action='version', version=version_str, help="Show the version number and exit")

    return parser.parse_args()


def run():
    """
    Main function to execute the command line interface for processing BIDS directories.
    """
    args = get_args()

    # Validate the arguments
    try:
        args = validate_args(args)
    except ValueError as e:
        print(Fore.RED + str(e) + Style.RESET_ALL)
        return
    
    pipeline = QCPipeline(args)
    pipeline.run()


if __name__ == "__main__":
    run()


from abc import ABC, abstractmethod

class LoggerService(ABC):
    """
    Abstract class for logging adapters.
    """
    @abstractmethod
    def info(self, message: str):
        """
        Log an info message.
        """
        pass

    @abstractmethod
    def error(self, message: str):
        """
        Log an error message.
        """
        pass

    @abstractmethod
    def debug(self, message: str):
        """
        Log a debug message.
        """
        pass

    @abstractmethod
    def warning(self, message: str):
        """
        Log a warning message.
        """
        pass
    


import os
from CPACqc import __version__

def validate_args(args):
    """
    Validate command-line arguments.
    """
    # Check if the BIDS directory exists
    if not os.path.exists(args.bids_dir):
        raise ValueError(f"BIDS directory '{args.bids_dir}' does not exist.")
    
    # Check if the QC output directory exists or create it
    if args.qc_dir:
        os.makedirs(args.qc_dir, exist_ok=True)
    
    # Check if the config file exists
    if args.config and not os.path.exists(args.config):
        raise ValueError(f"Config file '{args.config}' does not exist.")
    
    # Check if the subject list is valid
    if args.sub:
        for sub in args.sub:
            if not isinstance(sub, str):
                raise ValueError(f"Invalid subject label: {sub}. It should be a string.")
            if not sub.startswith("sub-"):
                raise ValueError(f"Invalid subject label format: {sub}. It should start with 'sub-'.")

    # Check if the number of processes is valid
    if args.n_procs <= 0:
        raise ValueError(f"Number of processes '{args.n_procs}' should be a positive integer.")

    
    # return the validated arguments
    return args




# from CPACqc.utils.tabler.custom_bids2table._b2t import bids2table
# import pandas as pd
# from CPACqc.services.tabler_service import TablerService

# class Bids2TableDetails(TablerService):
    
#     def __init__(self, base_dir, subjects=None, workers=8):
#         """
#         Initialize the Bids2TableDetails class.

#         Parameters:
#         - base_dir (str): The base directory for BIDS data.
#         - subjects (list): List of subjects to process. If None, all subjects will be processed.
#         """
#         self.base_dir = base_dir
#         self.subjects = subjects if subjects else []
#         self.workers = workers

#     def _parse_bids(self):
#         """
#         Parse the BIDS directory and return a DataFrame with the results.

#         Returns:
#         - pd.DataFrame: DataFrame containing the parsed BIDS data.
#         """
#         df = bids2table(self.base_dir, subject=self.subjects, workers = self.workers).flat
#         return df

#     def get_dataframe(self):
#         """
#         Convert the parsed BIDS data to a DataFrame.

#         Returns:
#         - pd.DataFrame: DataFrame containing the parsed BIDS data.
#         """
#         df = self._parse_bids()
#         return df



import bids2table as b2t
import pandas as pd
from CPACqc.services.tabler_service import TablerService

class Bids2TableDetails(TablerService):
    
    def __init__(self, base_dir, subjects=None, workers=8):
        """
        Initialize the Bids2TableDetails class.

        Parameters:
        - base_dir (str): The base directory for BIDS data.
        - subjects (list): List of subjects to process. If None, all subjects will be processed.
        """
        self.base_dir = base_dir
        self.subjects = subjects if subjects else []
        self.workers = workers
        self.arrow = self._parse_bids()

    def _parse_bids(self):
        return b2t.index_dataset(root=self.base_dir, include_subjects=self.subjects, show_progress=True)

    def get_dataframe(self):
        """
        Convert the parsed BIDS data to a DataFrame.

        Returns:
        - pd.DataFrame: DataFrame containing the parsed BIDS data.
        """
        df = self.arrow.to_pandas(types_mapper=pd.ArrowDtype)
        return df



    
    



from abc import ABC, abstractmethod
import pandas as pd

class TablerService(ABC):
    @abstractmethod
    def get_dataframe(self, base_dir: str, subjects: list, workers:int) -> pd.DataFrame:
        """
        Abstract method to convert a BIDS directory to a DataFrame.
        
        Parameters:
        - base_dir: str, path to the BIDS directory
        - subjects: list of str, subject IDs to include in the tabulation
        
        Returns:
        - df: pandas DataFrame, tabulated data
        """
        pass



import nibabel as nib
import numpy as np
from nilearn.plotting import plot_stat_map

def plot_nii_overlay(in_nii, plot_loc, background=None, volume=None, cmap='viridis', title=None, alpha=0.8, threshold=20):
    im = nib.load(in_nii)
    if volume is not None:
        v = volume
        im = nib.Nifti1Image(im.get_fdata()[:,:,:,v], header=im.header, affine=im.affine)

    if threshold != 'auto':
        lb = np.percentile(np.sort(np.unique(im.get_fdata())), float(threshold))
    else:
        lb = threshold

    if background and background != 'None':
        bg = nib.load(background)
        plot_stat_map(im, bg_img=bg, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))
    
    elif background == None or background == 'None':
        plot_stat_map(im, bg_img=None, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))
                                                                                                        
    else:
        plot_stat_map(im, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))


from CPACqc.utils.logging.log import FileLogger
from datetime import datetime
import os

log_dir = os.path.join(os.getcwd(), "logs")
os.makedirs(log_dir, exist_ok=True)
log_file = os.path.join(log_dir, f"cpacqc_{datetime.now().strftime('%Y%m%d_%H%M%S')}.log")

logger = FileLogger(name="CPACqc", filename=log_file)



from dataclasses import dataclass, field
from typing import Optional
import os
import pandas
from CPACqc.utils.logging.log import FileLogger as logger
from CPACqc.core.utils import *
import json


@dataclass
class Table:
    
    original_table:Optional[pandas.DataFrame]
    processed_table:Optional[pandas.DataFrame] = field(init=False)
    all_columns_old = ['dataset', 'dataset_type', 'dataset_path', 'dataset_description', 'sub',
       'ses', 'datatype', 'suffix', 'ext', 'extra_entities', 'sample', 'task',
       'tracksys', 'acq', 'nuc', 'voi', 'ce', 'trc', 'stain', 'rec', 'dir',
       'run', 'mod', 'echo', 'flip', 'inv', 'mt', 'part', 'proc', 'hemi',
       'space', 'split', 'recording', 'chunk', 'seg', 'res', 'den', 'label',
       'desc', 'json', 'file_path', 'link_target', 'mod_time', 'file_name',
       'resource_name', 'sub_ses', 'scan']
    all_columns_new = ['dataset', 'sub', 'ses', 'sample', 'task', 'tracksys', 'acq', 'nuc',
       'voi', 'ce', 'trc', 'stain', 'rec', 'dir', 'run', 'mod', 'echo', 'flip',
       'inv', 'mt', 'part', 'proc', 'hemi', 'space', 'split', 'recording',
       'chunk', 'seg', 'res', 'den', 'label', 'desc', 'datatype', 'suffix',
       'ext', 'extra_entities', 'root', 'path']

    columns_to_keep = ['sub', 'ses', 'file_path', 'file_name', 'resource_name', 'datatype', 'space', 'scan', 'json']

    def __post_init__(self):
        """
        Initialize the Table class and preprocess the original table.
        """
        if self.original_table is None:
            raise ValueError("original_table cannot be None")
        
        # Check if the original table is empty
        if self.original_table.empty:
            logger.error("The original table is empty.")
            raise ValueError("The original table is empty.")
        
        # Preprocess the original table
        self.preprocess()

        
    def preprocess(self):
        """
        Preprocess the original table to extract relevant information.
        """
        print(Fore.YELLOW + "Preprocessing the original table..." + Fore.RESET)
        df = self.original_table
        # Convert all dicts/lists in every column to string
        for col in df.columns:
            df[col] = df[col].apply(lambda x: str(x) if isinstance(x, (dict, list)) else x)
        df = df.fillna("")  # <-- Fill NAs before any comparison

        # Now safe to check for empty columns
        for col in df.columns:
            if df[col].nunique() == 1 and df[col].iloc[0] == "":
                df = df.drop(columns=[col])

        df = df.astype(str)

        files = ["nii.gz", ".nii"]

        # Filter rows where file_path ends with .nii.gz or .nii
        nii_gz_files = df[df.path.str.endswith(tuple(files))]

        # Filter rows and omit xfm.nii.gz files
        nii_gz_files = nii_gz_files.loc[~nii_gz_files.path.str.contains("xfm.nii.gz")]

        # Add a column that breaks the file_path to the last name of the file and drops extension
        nii_gz_files.loc[:, "file_name"] = nii_gz_files.path.apply(lambda x: os.path.basename(x).split(".")[0])
        # Combine sub and ses columns to create a new column called sub_ses
        nii_gz_files.loc[:, "sub_ses"] = nii_gz_files.apply(get_sub_ses, axis=1)
        nii_gz_files.loc[:, "resource_name"] = nii_gz_files.apply(lambda row: gen_resource_name(row), axis=1)

        # add root column and path column to save as file_path
        nii_gz_files.loc[:, "file_path"] = nii_gz_files.apply(lambda row: os.path.join(row.root, row.path), axis=1)
        nii_gz_files.loc[:, "json_path"] = nii_gz_files.apply(lambda row: os.path.join(row.root, row.path.replace(".nii.gz", ".json")), axis=1)
        nii_gz_files.loc[:, "json"] = nii_gz_files["json_path"].apply(safe_load_json)

        nii_gz_files = nii_gz_files[nii_gz_files.file_path.apply(lambda x: is_3d_or_4d(x))]

        # Check if the space column is empty and fill it accordingly
        nii_gz_files.loc[:, "space"] = nii_gz_files.apply(lambda x: fill_space(x), axis=1)

        # Create a new column called scan that combines task and run columns
        nii_gz_files.loc[:, "scan"] = nii_gz_files.apply(get_scan, axis=1)

        nii_gz_files.loc[:, "datatype"] = nii_gz_files.apply(lambda x: get_datatype(x), axis=1)

        self.processed_table = nii_gz_files

def get_datatype(row):
    datatype = (
        row["path"]
        .replace(".nii.gz", "")
        .replace(".nii", "")
        .replace(row["resource_name"], "")
        .replace(f"sub-{row['sub']}", "")
        .replace(f"ses-{row['ses']}", "")
        .replace(f"space-{row['space']}", "")
        .replace(row["scan"], "")
        .replace("_", "")
        .replace("-", "")
        .replace("/", "")
    )
    return datatype


    

def safe_load_json(json_path):
    if os.path.exists(json_path):
        try:
            with open(json_path, "r") as f:
                json_metadata = clean_and_flatten_json_column(json.load(f))
                return json_metadata
        except Exception as e:
            logger.error(f"Could not load JSON: {json_path} ({e})")
            return None
    else:
        return None

@dataclass
class Resource:
    """
    Class to represent a resource in the table.
    """
    sub: str
    ses: str
    file_path: str
    file_name: str
    resource_name: str
    datatype: str
    space: str
    scan: str
    json: Optional[str] = None

    
    


from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm
from colorama import Fore, Style
from typing import Callable, Iterable, Any, List

from CPACqc.services.multiprocessing_service import MultiprocessingService
from CPACqc.core.logger import logger

class Multiprocess(MultiprocessingService):
    def __init__(self, func, args, n_procs):
        self.func = func
        self.args = args
        self.n_procs = n_procs
        self.not_plotted = []

    def run(self) -> List[Any]:
        with ProcessPoolExecutor(max_workers=self.n_procs) as executor:
            futures = {executor.submit(self.func, arg): arg for arg in self.args}
            for future in tqdm(as_completed(futures), total=len(futures), desc="Processing ..."):
                try:
                    result = future.result()
                    logger.info(f"Successfully processed {futures[future]}: {result}")
                except Exception as e:
                    if "terminated abruptly" in str(e):
                        print(Fore.RED + f"Error processing {futures[future]}: {e}\n Try with lower number of processes" + Style.RESET_ALL)
                    self.not_plotted.append(futures[future])
        return self.not_plotted



import pytest
import os
from unittest.mock import MagicMock, patch
from CPACqc.core.validation import validate_args

@patch("os.path.exists")
@patch("os.makedirs")
def test_validate_args_valid(mock_makedirs, mock_exists):
    # Setup mocks
    mock_exists.side_effect = lambda path: path in ["/bids", "config.csv"]
    args = MagicMock()
    args.bids_dir = "/bids"
    args.qc_dir = "/qc"
    args.config = "config.csv"
    args.sub = ["sub-01", "sub-02"]
    args.n_procs = 2
    args.version = False

    result = validate_args(args)
    assert result == args
    mock_makedirs.assert_called_with("/qc", exist_ok=True)

@patch("os.path.exists")
def test_validate_args_missing_bids(mock_exists):
    mock_exists.return_value = False
    args = MagicMock()
    args.bids_dir = "/missing"
    args.qc_dir = None
    args.config = None
    args.sub = None
    args.n_procs = 1
    args.version = False

    with pytest.raises(ValueError, match="BIDS directory '/missing' does not exist."):
        validate_args(args)

@patch("os.path.exists")
def test_validate_args_invalid_config(mock_exists):
    def exists_side_effect(path):
        return path == "/bids"
    mock_exists.side_effect = exists_side_effect
    args = MagicMock()
    args.bids_dir = "/bids"
    args.qc_dir = None
    args.config = "missing.csv"
    args.sub = None
    args.n_procs = 1
    args.version = False

    with pytest.raises(ValueError, match="Config file 'missing.csv' does not exist."):
        validate_args(args)

@patch("os.path.exists", return_value=True)
def test_validate_args_invalid_subject_label(mock_exists):
    args = MagicMock()
    args.bids_dir = "/bids"
    args.qc_dir = None
    args.config = None
    args.sub = ["01"]
    args.n_procs = 1
    args.version = False

    with pytest.raises(ValueError, match="Invalid subject label format: 01. It should start with 'sub-'"):
        validate_args(args)

@patch("os.path.exists", return_value=True)
def test_validate_args_invalid_n_procs(mock_exists):
    args = MagicMock()
    args.bids_dir = "/bids"
    args.qc_dir = None
    args.config = None
    args.sub = None
    args.n_procs = 0
    args.version = False

    with pytest.raises(ValueError, match="Number of processes '0' should be a positive integer."):
        validate_args(args)

@patch("os.path.exists", return_value=True)
def test_validate_args_version_flag(mock_exists, capsys):
    args = MagicMock()
    args.bids_dir = "/bids"
    args.qc_dir = None
    args.config = None
    args.sub = None
    args.n_procs = 1
    args.version = True

    from CPACqc import __version__
    validate_args(args)
    captured = capsys.readouterr()
    assert f"CPACqc version : {__version__}" in captured.out



from abc import ABC, abstractmethod

class ReportGeneratorService(ABC):
    """
    Abstract base class for report generation services.
    """

    @abstractmethod
    def generate_report(self, entity, config):
        """
        Generate a report based on the provided data.

        :param data: The data to be included in the report.
        :return: The generated report.
        """
        pass

    @abstractmethod
    def save_report(self, report, file_path):
        """
        Save the generated report to a file.

        :param report: The report to be saved.
        :param file_path: The path where the report will be saved.
        """
        pass


import pandas as pd
from multiprocessing import Pool
import os
from tqdm import tqdm
from colorama import Fore, Style, init
import nibabel as nib
from nilearn.plotting import plot_stat_map
import matplotlib.pyplot as plt
import numpy as np

from CPACqc.utils.plotting.nii_plotter import plot_nii_overlay
from CPACqc.services.plotting_service import PlottingService

class Plot(PlottingService):
    def __init__(self, sub, ses, file_path_1, file_path_2, file_name, plots_dir, plot_path):
        self.sub = sub
        self.ses = ses
        self.file_path_1 = file_path_1
        self.file_path_2 = file_path_2
        self.file_name = file_name
        self.plots_dir = plots_dir
        self.plot_path = plot_path
        self.dim = len(nib.load(file_path_1).shape)
        self.volume_index = 0 if self.dim == 4 else None
        
    def plot_overlay(self):
        
        try:
            plot_nii_overlay(
                self.file_path_1,
                self.plot_path,
                background=self.file_path_2 if self.file_path_2 else None,
                volume=self.volume_index,
                cmap='bwr',
                title="",
                alpha=0.5,
                threshold="auto"
            )
        except Exception as e:
            # print(Fore.RED + f"Error on {file_name}" + Style.RESET_ALL)
            # print(Fore.RED + f"Error: {e}" + Style.RESET_ALL)
            return f"Error on {file_name}: {e}"
        return f"Successfully plotted"

    def __repr__(self):
        return f"{self.file_name}"


import pytest
from unittest.mock import patch, MagicMock
from CPACqc.services.multiprocessing_service import MultiprocessingPort
from CPACqc.utils.logging.log import FileLogger as logger

from CPACqc.utils.multiprocessing.multiprocessing_utils import ProcessPoolMultiprocessing

def dummy_func(x):
    return x * 2

def dummy_func_error(x):
    raise ValueError("Test error")

def test_run_success(monkeypatch):
    # Patch logger to avoid actual logging
    monkeypatch.setattr(logger, "info", lambda msg: None)
    monkeypatch.setattr(logger, "error", lambda msg: None)

    runner = ProcessPoolMultiprocessing()
    args = [1, 2, 3]
    not_plotted = runner.run(dummy_func, args, n_procs=2)
    assert not not_plotted  # Should be empty if all succeed

def test_run_with_error(monkeypatch):
    monkeypatch.setattr(logger, "info", lambda msg: None)
    monkeypatch.setattr(logger, "error", lambda msg: None)

    runner = ProcessPoolMultiprocessing()
    args = [1, 2]
    not_plotted = runner.run(dummy_func_error, args, n_procs=2)
    assert set(not_plotted) == set(args)